/* ======================
 * Eastwood Chart Servlet
 * ======================
 *
 * (C) Copyright 2007, by Object Refinery Limited.
 *
 * Project Info:  http://www.jfree.org/eastwood/index.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc. 
 * in the United States and other countries.]
 *
 * -----------------
 * ChartServlet.java
 * -----------------
 * (C) Copyright 2007, by Object Refinery Limited.
 *
 * Original Author:  David Gilbert (for Object Refinery Limited);
 *
 * Changes
 * -------
 * 13-Dec-2007 : Version 1 (DG);
 *
 */

package org.jfree.eastwood;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Stroke;
import java.io.IOException;
import java.io.OutputStream;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.AxisLocation;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.IntervalMarker;
import org.jfree.chart.plot.PiePlot;
import org.jfree.chart.plot.PiePlot3D;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.AbstractRenderer;
import org.jfree.chart.renderer.category.BarRenderer;
import org.jfree.chart.renderer.category.StackedBarRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.title.Title;
import org.jfree.data.Range;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.data.general.DefaultPieDataset;
import org.jfree.data.general.PieDataset;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.Layer;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.RectangleInsets;

/**
 * A servlet that returns a chart as a PNG image.  The servlet should
 * accept arguments formatted according to the Google Chart API.
 */
public class ChartServlet extends HttpServlet {

    /**
     * Default constructor.
     */
    public ChartServlet() {
        // nothing required
    }

    /**
     * Process a GET request.
     *
     * @param request  the request.
     * @param response  the response.
     *
     * @throws ServletException if there is a servlet related problem.
     * @throws IOException if there is an I/O problem.
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException {

        List axes = new java.util.ArrayList();
        
        OutputStream out = response.getOutputStream();
        try {
            
            // *** CHART TYPE **
            String chartType = request.getParameter("cht");
            JFreeChart chart = null;
            int dataType = -1;  // 0 = PieDataset; 1 = XYDataset; 2=special
            
            // pie charts: 'p' and 'p3'
            if (chartType.equals("p")) {
                chart = createPieChart();
                dataType = 0;
            }
            else if (chartType.equals("p3")) {
                chart = createPieChart3D();
                dataType = 0;
            }
            // line chart: 'lc'
            else if (chartType.equals("lc")) {
                chart = createLineChart();
                dataType = 1;
            }
            // xy chart: 'lxy'
            else if (chartType.equals("lxy")) {
                chart = createLineChart();
                dataType = 3;
            }
            // bar charts: 'bhs', 'bhg', 'bhs' and 'bhg'
            else if (chartType.equals("bhs")) {
                chart = createStackedBarChart(PlotOrientation.HORIZONTAL);
                dataType = 2;
            }
            else if (chartType.equals("bhg")) {
                chart = createBarChart(PlotOrientation.HORIZONTAL);
                dataType = 2;                
            }
            else if (chartType.equals("bvs")) { 
                chart = createStackedBarChart(PlotOrientation.VERTICAL);
                dataType = 2;                
            }
            else if (chartType.equals("bvg")) {
                chart = createBarChart(PlotOrientation.VERTICAL);
                dataType = 2;                
            }
            // scatter chart: 's'
            else if (chartType.equals("s")) {
                chart = createScatterChart();
                dataType = 4;
            }
            else if (chartType.equals("v")) {
                throw new RuntimeException("Venn diagrams not implemented.");
                // TODO: fix this.
            }
            else {
                throw new ServletException("Unknown chart type: " + chartType);
            }
      
            
            // *** CHART AXES ***
            String axisStr = request.getParameter("chxt");
            if (axisStr != null) {
                if (chart.getPlot() instanceof XYPlot) {
                    XYPlot plot = (XYPlot) chart.getPlot();
                    processAxisStr(plot, axisStr, axes);
                }
                else if (chart.getPlot() instanceof CategoryPlot) {
                    CategoryPlot plot = (CategoryPlot) chart.getPlot();
                    if (plot.getOrientation() == PlotOrientation.VERTICAL) {
                        processAxisStrV(plot, axisStr, axes);
                    }
                    else {
                        processAxisStrH(plot, axisStr, axes);
                    }
                }
            }
            
            
            // *** AXIS RANGES ***
            String axisRangeStr = request.getParameter("chxr");
            if (axisRangeStr != null) {
                String[] ranges = breakString(axisRangeStr, '|');
                for (int i = 0; i < ranges.length; i++) {
                    int comma1 = ranges[i].indexOf(',');
                    int comma2 = ranges[i].indexOf(',', comma1 + 1);
                    int axisIndex = Integer.parseInt(ranges[i].substring(0, 
                            comma1));
                    float lowerBound = Float.parseFloat(ranges[i].substring(
                            comma1 + 1, comma2));
                    float upperBound = Float.parseFloat(ranges[i].substring(
                            comma2 + 1));
                    GValueAxis axis = (GValueAxis) axes.get(axisIndex);
                    axis.setRange(lowerBound, upperBound);
                }
            }
            
            
            // *** AXIS LABELS ***
            String axisLabelStr = request.getParameter("chxl");
            if (axisLabelStr != null) {
                Pattern p = Pattern.compile("\\d+:\\|");
                Matcher m = p.matcher(axisLabelStr);
                if (m.find()) {
                    int keyStart = m.start();
                    int labelStart = m.end();
                    while (m.find(labelStart)) {
                        String keyStr = axisLabelStr.substring(keyStart, 
                                labelStart - 2);
                        int axisIndex = Integer.parseInt(keyStr);
                        keyStart = m.start();
                        String labelStr = axisLabelStr.substring(labelStart, 
                                keyStart - 1);
                        String[] labels = breakString(labelStr, '|');
                        GLabelledAxis axis = (GLabelledAxis) axes.get(axisIndex);
                        axis.setTickLabels(Arrays.asList(labels));
                        labelStart = m.end();
                    }
                    // process the final item
                    String keyStr = axisLabelStr.substring(keyStart, 
                            labelStart - 2);
                    String labelStr = axisLabelStr.substring(labelStart);
                    int axisIndex = Integer.parseInt(keyStr);                    
                    if (labelStr.endsWith("|")) {
                        labelStr = labelStr.substring(0, labelStr.length() - 1);
                    }
                    String[] labels = breakString(labelStr, '|');
                    GLabelledAxis axis = (GLabelledAxis) axes.get(axisIndex);
                    axis.setTickLabels(Arrays.asList(labels));
                    
                }
                else {
                    throw new RuntimeException("No matching pattern!");
                }
                
            }
            
            
            // ** EXPLICIT AXIS LABEL POSITIONS
            String axisPositionStr = request.getParameter("chxp");
            if (axisPositionStr != null) {
                String[] positions = breakString(axisPositionStr, '|');
                for (int i = 0; i < positions.length; i++) {
                    int c1 = positions[i].indexOf(',');
                    int axisIndex = Integer.parseInt(positions[i].substring(0, c1));
                    String remainingStr = positions[i].substring(c1 + 1);
                    String[] valueStr = breakString(remainingStr, ',');
                    List tickValues = new java.util.ArrayList(valueStr.length);
                    GValueAxis axis = (GValueAxis) axes.get(axisIndex);
                    for (int j = 0; j < valueStr.length; j++) {
                        tickValues.add(new Float(Float.parseFloat(valueStr[j])));
                    }
                    axis.setTickValues(tickValues);
                }
            }
            
            
            // *** CHART TITLE ***
            String titleStr = request.getParameter("chtt");
            if (titleStr != null) {
                // process the title
                String[] s = breakString(titleStr, '|');
                for (int i = 0; i < s.length; i++) {
                    TextTitle t = new TextTitle(s[i].replace('+', ' '));
                    t.setPaint(Color.gray);
                    chart.addSubtitle(t);
                }
                // and the font and colour
                String fontStr = request.getParameter("chts");
                if (fontStr != null) {
                    int c1 = fontStr.indexOf(',');
                    String colorStr = null;
                    String fontSizeStr = null;
                    if (c1 != -1) {
                        colorStr = fontStr.substring(0, c1);
                        fontSizeStr = fontStr.substring(c1 + 1);
                    }
                    else {
                        colorStr = fontStr;
                    }
                    Color color = parseColor(colorStr);
                    int size = 12;
                    if (fontSizeStr != null) {
                        size = Integer.parseInt(fontSizeStr);
                    }
                    for (int i = 0; i < chart.getSubtitleCount(); i++) {
                        Title t = chart.getSubtitle(i);
                        if (t instanceof TextTitle) {
                            TextTitle tt = (TextTitle) t;
                            tt.setPaint(color);
                            Font f = tt.getFont();
                            tt.setFont(new Font(f.getFontName(), f.getStyle(), 
                                    size));
                        }
                    }
                }
            }
            
            // *** CHART DATA ***
            String dataStr = request.getParameter("chd");
            
            // 'p' and 'p3' - create PieDataset
            if (dataType == 0) {
                PieDataset dataset = parsePieDataset(dataStr);
                PiePlot plot = (PiePlot) chart.getPlot();
                plot.setDataset(dataset);
            }
            
            // 'lc' - create XYDataset
            else if (dataType == 1) {
                XYPlot plot = (XYPlot) chart.getPlot();
                ValueAxis xAxis = plot.getDomainAxis();
                ValueAxis yAxis = plot.getRangeAxis();
                XYDataset dataset = parseXYDataset(dataStr, xAxis.getRange(), 
                        yAxis.getRange());
                plot.setDataset(dataset);
            }
            
            // 'bhs', 'bhg', 'bvs', 'bvg' - create specialised TableXYDataset
            else if (dataType == 2) {
                CategoryPlot plot = (CategoryPlot) chart.getPlot();
                ValueAxis yAxis = plot.getRangeAxis();
                CategoryDataset dataset = parseCategoryDataset(dataStr, 
                        yAxis.getRange());
                plot.setDataset(dataset);                
            }
            
            // 'lxy'
            else if (dataType == 3) {
                XYPlot plot = (XYPlot) chart.getPlot();
                ValueAxis xAxis = plot.getDomainAxis();
                ValueAxis yAxis = plot.getRangeAxis();
                XYDataset dataset = parseXYDataset2(dataStr, xAxis.getRange(), 
                        yAxis.getRange());
                plot.setDataset(dataset);                
            }
            else if (dataType == 4) {
                XYPlot plot = (XYPlot) chart.getPlot();
                ValueAxis xAxis = plot.getDomainAxis();
                ValueAxis yAxis = plot.getRangeAxis();
                XYSeriesCollection dataset = parseScatterDataset(dataStr, 
                        xAxis.getRange(), yAxis.getRange());
                if (dataset.getSeriesCount() > 1) {
                    dataset.removeSeries(1);
                }
                plot.setDataset(dataset);                
            }
            
            
            // *** CHART COLOURS ***
            String colorStr = request.getParameter("chco");
            if (colorStr != null) {
                Color[] colors = parseColors(colorStr);
                if (dataType == 0) {
                    PiePlot plot = (PiePlot) chart.getPlot();
                    applyColorsToPiePlot(plot, colors);
                }
                else {
                    AbstractRenderer renderer = null;
                    if (chart.getPlot() instanceof CategoryPlot) {
                        CategoryPlot plot = (CategoryPlot) chart.getPlot();
                        renderer = (AbstractRenderer) plot.getRenderer();
                        renderer.setBasePaint(colors[0]);
                    }
                    else if (chart.getPlot() instanceof XYPlot) {
                        XYPlot plot = (XYPlot) chart.getPlot();
                        renderer = (AbstractRenderer) plot.getRenderer();
                        renderer.setBasePaint(colors[colors.length - 1]);
                    }
                    for (int i = 0; i < colors.length; i++) {
                        renderer.setSeriesPaint(i, colors[i]);
                    }
                }
            }
            else {
                Plot plot = chart.getPlot();
                if (plot instanceof PiePlot) {
                    applyColorsToPiePlot((PiePlot) chart.getPlot(), 
                            new Color[] { new Color(255, 153, 0) });
                }
            }
            
            
            // *** CHART LINE STYLES ***
            String lineStr = request.getParameter("chls");
            if (lineStr != null) {
                Stroke[] strokes = parseLineStyles(lineStr);
                XYPlot plot = (XYPlot) chart.getPlot();
                XYItemRenderer renderer = plot.getRenderer();
                for (int i = 0; i < strokes.length; i++) {
                    renderer.setSeriesStroke(i, strokes[i]);
                }
                renderer.setBaseStroke(strokes[strokes.length - 1]);
            }
            
            
            // *** CHART GRID LINES
            if (dataType != 0) {
                String gridStr = request.getParameter("chg");
                if (gridStr != null) {
                    // FIXME: read the grid line style
                    Plot p = chart.getPlot();
                    if (p instanceof CategoryPlot) {
                        CategoryPlot plot = (CategoryPlot) p;
                        plot.setDomainGridlinesVisible(true);
                        plot.setRangeGridlinesVisible(true);
                    }
                    else if (p instanceof XYPlot) {
                        XYPlot plot = (XYPlot) p;
                        plot.setDomainGridlinesVisible(true);
                        plot.setRangeGridlinesVisible(true);                        
                    }
                }
            }
          
            // *** CHART LEGEND LABELS
            if (dataType == 0) {  // pie chart
                String labelStr = request.getParameter("chl");
                if (labelStr != null) {
                    String[] s = breakString(labelStr, '|');                    
                    List labels = Arrays.asList(s);
                    PiePlot plot = (PiePlot) chart.getPlot();
                    if (labels.size() > 0) {
                        plot.setLabelGenerator(
                                new GPieSectionLabelGenerator(labels));
                    }
                }
            }
            else {
                // FIXME: Handle CategoryPlot
                String legendStr = request.getParameter("chdl");
                if (legendStr != null) {
                    // process the title
                    String[] s = breakString(legendStr, '|');
                    List labels = Arrays.asList(s);
                    if (labels.size() > 0) {
                        XYPlot plot = (XYPlot) chart.getPlot();
                        XYItemRenderer renderer = plot.getRenderer();
                        renderer.setLegendItemLabelGenerator(
                                new GXYSeriesLabelGenerator(labels));
                        renderer.setBaseSeriesVisibleInLegend(false);
                        for (int i = 0; i < labels.size(); i++) {
                            renderer.setSeriesVisibleInLegend(i, Boolean.TRUE);
                        }
                        LegendTitle legend = new LegendTitle(chart.getPlot());
                        legend.setPosition(RectangleEdge.RIGHT);
                        chart.addSubtitle(legend);
                    }
                }
            }
            
            
            // *** CHART MARKERS ***
            String markerStr = request.getParameter("chm");
            if (markerStr != null) {
                String[] markers = breakString(markerStr, '|');
                for (int i = 0; i < markers.length; i++) {
                    System.out.println(markers[i]);
                    addMarker(markers[i], chart);
                }
            }
            
            
            // *** CHART FILL ***/
            String fillStr = request.getParameter("chf");
            if (fillStr != null) {
                // process the 1 or 2 fill specs
                int i = fillStr.indexOf('|');
                if (i == -1) {
                    processFillSpec(fillStr, chart);
                }
                else {
                    String fs1 = fillStr.substring(0, i);
                    String fs2 = fillStr.substring(i + 1);
                    processFillSpec(fs1, chart);
                    processFillSpec(fs2, chart);
                }
            }
           
            
            // *** CHART SIZE ***
            String size = request.getParameter("chs");
            int[] dims = new int[2];
            if (size != null) {
                dims = parseChartDimensions(size);
            }
            else {
                dims = new int[] {200, 125};
            }
            if (chart != null) {
                response.setContentType("image/png");
                ChartUtilities.writeChartAsPNG(out, chart, dims[0], dims[1]);
            }
            
        }
        catch (Exception e) {
            System.err.println(e.toString());
        }
        finally {
            out.close();
        }

    }
    
    /**
     * A utility method that splits the supplied string at every occurrence of
     * the specified separator.
     * 
     * @param text  the text (<code>null</code> not permitted).
     * @param separator  the separator character.
     * 
     * @return An array of strings.
     */
    private String[] breakString(String text, char separator) {
        if (text == null) {
            throw new IllegalArgumentException("Null 'text' argument.");
        }
        List temp = new java.util.ArrayList();
        String current = text;
        int i = current.indexOf(separator);
        while (i != -1) {
            String line = current.substring(0, i);
            temp.add(line);
            current = current.substring(i + 1);
            i = current.indexOf(separator);
        }
        temp.add(current);
        String[] result = new String[temp.size()];
        for (int j = 0; j < temp.size(); j++) {
            result[j] = (String) temp.get(j);
        }
        return result;
    }
    
    /**
     * Adds a marker to the chart.
     * 
     * @param markerStr  the string encoding of the marker.
     * @param chart  the chart to apply the marker to.
     */
    private void addMarker(String markerStr, JFreeChart chart) {
        String[] args = breakString(markerStr, ',');
        Plot p = chart.getPlot();
        Color c = parseColor(args[1]);
        if ("r".equals(args[0])) {
            // add a range marker
            float f0 = Float.parseFloat(args[3]);
            float f1 = Float.parseFloat(args[4]);
            if (p instanceof CategoryPlot) {
                CategoryPlot cp = (CategoryPlot) p;
                Range yRange = cp.getRangeAxis().getRange();
                double r0 = yRange.getLowerBound();
                double rl = yRange.getLength();
                IntervalMarker m = new IntervalMarker(r0 + Math.min(f0, f1) 
                        * rl, r0 + Math.max(f0, f1) * rl);
                m.setPaint(c);
                cp.addRangeMarker(m, Layer.BACKGROUND);
            }
            else if (p instanceof XYPlot) {
                XYPlot cp = (XYPlot) p;
                Range yRange = cp.getRangeAxis().getRange();
                double r0 = yRange.getLowerBound();
                double rl = yRange.getLength();
                IntervalMarker m = new IntervalMarker(r0 + Math.min(f0, f1) 
                        * rl, r0 + Math.max(f0, f1) * rl);
                m.setPaint(c);
                cp.addRangeMarker(m, Layer.BACKGROUND);
            }
            
        }
        else if ("R".equals(args[0])) {
            // add a domain marker
            float f0 = Float.parseFloat(args[3]);
            float f1 = Float.parseFloat(args[4]);
            if (p instanceof XYPlot) {
                XYPlot xyp = (XYPlot) p;
                Range yRange = xyp.getRangeAxis().getRange();
                double r0 = yRange.getLowerBound();
                double rl = yRange.getLength();
                IntervalMarker m = new IntervalMarker(r0 + Math.min(f0, f1) 
                        * rl, r0 + Math.max(f0, f1) * rl);
                m.setPaint(c);
                xyp.addDomainMarker(m, Layer.BACKGROUND);
            }
        }
    }
    
    /**
     * Creates a color from the supplied string, which should be in RRGGBB 
     * format, or RRGGBBAA.
     * 
     * @param text  the text encoding (<code>null</code> not permitted).
     * 
     * @return A color.
     */
    private Color parseColor(String text) {
        if (text == null) {    
            throw new IllegalArgumentException(
                    "Null 'text' argument (in parseColor(String)).");
        }
        return Color.decode("0x" + text);
    }
    
    /**
     * Parses a string containing a comma-separated list of color values (in
     * the format RRGGBB or RRGGBBAA).
     * 
     * @param text  the text (<code>null</code> not permitted).
     * 
     * @return The colors.
     */
    private Color[] parseColors(String text) {
        if (text == null) {    
            throw new IllegalArgumentException(
                    "Null 'text' argument (in parseColors(String)).");
        }
        String[] codes = breakString(text, ',');
        Color[] result = new Color[codes.length];
        for (int i = 0; i < codes.length; i++) {
            if (codes[i].length() > 0) {
                result[i] = Color.decode("0x" + codes[i]);
            }
            else {
                result[i] = Color.black;
            }
        }
        return result;
    }
    
    private void applyColorsToPiePlot(PiePlot plot, Color[] colors) {
        if (colors.length == 1) {
            Color c = colors[0];
            colors = new Color[2];
            colors[0] = c;
            colors[1] = new Color(255 - ((255 - c.getRed()) / 5), 
                    255 - ((255 - c.getGreen()) / 5), 
                    255 - ((255 - c.getBlue()) / 5));
        }
        PieDataset dataset = plot.getDataset();
        int sectionCount = dataset.getItemCount();
        if (colors.length < sectionCount) {  // we need to interpolate some
                                             // colors
            for (int i = 0; i < colors.length - 1; i++) {
                Color c1 = colors[i];
                Color c2 = colors[i + 1];
                int s1 = sectionIndexForColor(i, colors.length, sectionCount);
                int s2 = sectionIndexForColor(i + 1, colors.length, 
                        sectionCount);
                for (int s = s1; s <= s2; s++) {
                    Color c = interpolatedColor(c1, c2, s - s1, s2 - s1);
                    plot.setSectionPaint(dataset.getKey(s), c);
                }
            }
        }
        else {
            for (int i = 0; i < sectionCount; i++) {
                plot.setSectionPaint(dataset.getKey(i), colors[i]);
            }
        }
    }
    

    private int sectionIndexForColor(int item, int itemCount, 
            int sectionCount) {
        return (int) (Math.min(sectionCount - 1, item * sectionCount 
                / (itemCount - 1.0)));
    }
    
    private Color interpolatedColor(Color c1, Color c2, int index, int range) {
        if (index == 0) {
            return c1;
        }
        if (index == range) {
            return c2;
        }
        double fraction = (index + 0.0) / range;
        int r1 = c1.getRed(); int g1 = c1.getGreen(); int b1 = c1.getBlue();
        int r2 = c2.getRed(); int g2 = c2.getGreen(); int b2 = c2.getBlue();
        return new Color((int) (r1 + fraction * (r2 - r1)), 
                (int) (g1 + fraction * (g2 - g1)), 
                (int) (b1 + fraction * (b2 - b1)));
    }
    
    /**
     * Parses a string containing a sequence of line style specifications,
     * returning an array of <code>Stroke</code> objects.
     * 
     * @param text  the line styles.
     * 
     * @return The strokes representing the line styles.
     */
    private Stroke[] parseLineStyles(String text) {
        if (text == null) {    
            throw new IllegalArgumentException(
                    "Null 'text' argument (in parseStrokes(String)).");
        }
        String[] codes = breakString(text, '|');
        Stroke[] result = new BasicStroke[codes.length];
        for (int i = 0; i < codes.length; i++) {
            float width = 1.0f;
            float lineRun = 1.0f;
            float gapRun = 0.0f;
            int pos = codes[i].indexOf(',');
            if (pos == -1) {
                width = Float.parseFloat(codes[i]);
            }
            else {
                String widthStr = codes[i].substring(0, pos);
                width = Float.parseFloat(widthStr);
                String remaining = codes[i].substring(pos + 1);
                pos = remaining.indexOf(',');
                if (pos == -1) {
                    lineRun = Float.parseFloat(remaining);
                    gapRun = lineRun;
                }
                else {
                    String s1 = remaining.substring(0, pos);
                    lineRun = Float.parseFloat(s1);
                    String s2 = remaining.substring(pos + 1);
                    gapRun = Float.parseFloat(s2);
                }
            }
            if (gapRun <= 0.0f) {
                result[i] = new BasicStroke(width, BasicStroke.CAP_ROUND, 
                        BasicStroke.JOIN_ROUND);   
            }
            else {
                result[i] = new BasicStroke(width, BasicStroke.CAP_BUTT, 
                        BasicStroke.JOIN_ROUND, 5.0f, 
                        new float[] {lineRun, gapRun}, 0.0f);
            }
        }
        return result;
    }
    
    /**
     * Process the string which contains an encoding for the chart background
     * fill, and apply it to the chart.
     * 
     * @param spec  the fill specification.
     * @param chart  the chart.
     */
    private void processFillSpec(String spec, JFreeChart chart) {
        if (spec.startsWith("bg")) {
            // do the chart background
            spec = spec.substring(3);
            if (spec.startsWith("s")) {
                spec = spec.substring(2);
                Color c = parseColor(spec);
                chart.setBackgroundPaint(c);
            }
            else if (spec.startsWith("lg")) {
                spec = spec.substring(3);
                String[] args = breakString(spec, ',');
                int angle = Integer.parseInt(args[0]);
                Color c0 = parseColor(args[1]);
                float f0 = Float.parseFloat(args[2]);
                Color c1 = parseColor(args[3]);
                float f1 = Float.parseFloat(args[4]);
                if (chart.getPlot() instanceof GXYPlot) {
                    GXYPlot gxyplot= (GXYPlot) chart.getPlot();
                    gxyplot.setF0(f0);
                    gxyplot.setF1(f1);
                    gxyplot.setAngle(Math.PI / 180.0 * angle);
                    gxyplot.setBackgroundPaint(new GradientPaint(0.0f, 0.0f, 
                            c0, 0.0f, 0.0f, c1));
                }
            }
        }
        else if (spec.startsWith("c")) {
            // do the plot background
            spec = spec.substring(2);
            if (spec.startsWith("s")) {
                spec = spec.substring(2);
                Color c = parseColor(spec);
                chart.getPlot().setBackgroundPaint(c);
            }
            else if (spec.startsWith("lg")) {
                spec = spec.substring(3);
                String[] args = breakString(spec, ',');
                int angle = Integer.parseInt(args[0]);
                Color c0 = parseColor(args[1]);
                float f0 = Float.parseFloat(args[2]);
                Color c1 = parseColor(args[3]);
                float f1 = Float.parseFloat(args[4]);
                if (chart.getPlot() instanceof GXYPlot) {
                    GXYPlot gxyplot= (GXYPlot) chart.getPlot();
                    gxyplot.setF0(f0);
                    gxyplot.setF1(f1);
                    gxyplot.setAngle(Math.PI / 180.0 * angle);
                    gxyplot.setBackgroundPaint(new GradientPaint(0.0f, 0.0f, 
                            c0, 0.0f, 0.0f, c1));
                }
            }
            else {
                throw new RuntimeException(
                        "'c' background fill not implemented yet.");
            }
        }
        else {
            throw new RuntimeException("Bad fill specification: " + spec);
        }
    }
    
    /**
     * Processes a string that indicates the axes that should be visible on 
     * the plot.
     * 
     * @param plot  the plot.
     * @param axisStr  the axis specification.
     * @param axes  a list that will be populated with any axes added to the
     *              plot.
     */
    private void processAxisStr(XYPlot plot, String axisStr, List axes) {
        int xAxisCount = 0;
        int yAxisCount = 0;
        for (int i = 0; i < axisStr.length(); i++) {
            char c = axisStr.charAt(i);
            if (c == 'x') {
                if (xAxisCount == 0) {
                    plot.getDomainAxis().setVisible(true);
                    axes.add(plot.getDomainAxis());
                    xAxisCount++;
                } 
                else {
                    GValueAxis axis = new GValueAxis();
                    axis.setAxisLineVisible(false);
                    plot.setDomainAxis(xAxisCount, axis);
                    plot.setDomainAxisLocation(xAxisCount, 
                            AxisLocation.BOTTOM_OR_LEFT);
                    axes.add(axis);
                    xAxisCount++;
                }
            }
            else if (c == 'y') {
                if (yAxisCount == 0) {
                    plot.getRangeAxis().setVisible(true);
                    axes.add(plot.getRangeAxis());
                    yAxisCount++;
                }
                else {
                    GValueAxis axis = new GValueAxis();
                    axis.setAxisLineVisible(false);
                    plot.setRangeAxis(yAxisCount, axis);
                    plot.setRangeAxisLocation(yAxisCount, 
                            AxisLocation.BOTTOM_OR_LEFT);
                    axes.add(axis);
                    yAxisCount++;
                }
            }
            else if (c == 'r') {
                GValueAxis axis = new GValueAxis();
                plot.setRangeAxis(yAxisCount, axis);
                plot.setRangeAxisLocation(yAxisCount, 
                        AxisLocation.BOTTOM_OR_RIGHT);
                axes.add(axis);
                yAxisCount++;
            }
            else if (c == 't') {
                GValueAxis axis = new GValueAxis();
                plot.setDomainAxis(xAxisCount, axis);
                plot.setDomainAxisLocation(xAxisCount, 
                        AxisLocation.TOP_OR_LEFT);
                axes.add(axis);
                xAxisCount++;
            }
            else if (c == ',') {
                // nothing to do
            }
            else {
                throw new RuntimeException("Bad character " + c);
            }
        }
        
    }
    
    /**
     * Processes a string that indicates the axes that should be visible on 
     * the plot.
     * 
     * @param plot  the plot.
     * @param axisStr  the axis specification.
     * @param axes  a list that will be populated with any axes added to the
     *              plot.
     */
    private void processAxisStrV(CategoryPlot plot, String axisStr, List axes) {
        int xAxisCount = 0;
        int yAxisCount = 0;
        for (int i = 0; i < axisStr.length(); i++) {
            char c = axisStr.charAt(i);
            if (c == 'x') {
                if (xAxisCount == 0) {
                    plot.getDomainAxis().setVisible(true);
                    axes.add(plot.getDomainAxis());
                    xAxisCount++;
                } 
                else {
                    GCategoryAxis axis = new GCategoryAxis();
                    axis.setAxisLineVisible(false);
                    plot.setDomainAxis(xAxisCount, axis);
                    plot.setDomainAxisLocation(xAxisCount, 
                            AxisLocation.BOTTOM_OR_LEFT);
                    axes.add(axis);
                    xAxisCount++;
                }
            }
            else if (c == 'y') {
                if (yAxisCount == 0) {
                    plot.getRangeAxis().setVisible(true);
                    axes.add(plot.getRangeAxis());
                    yAxisCount++;
                }
                else {
                    GValueAxis axis = new GValueAxis();
                    axis.setAxisLineVisible(false);
                    plot.setRangeAxis(yAxisCount, axis);
                    plot.setRangeAxisLocation(yAxisCount, 
                            AxisLocation.BOTTOM_OR_LEFT);
                    axes.add(axis);
                    yAxisCount++;
                }
            }
            else if (c == 'r') {
                GValueAxis axis = new GValueAxis();
                plot.setRangeAxis(yAxisCount, axis);
                plot.setRangeAxisLocation(yAxisCount, 
                        AxisLocation.BOTTOM_OR_RIGHT);
                axes.add(axis);
                yAxisCount++;
            }
            else if (c == 't') {
                GCategoryAxis axis = new GCategoryAxis();
                plot.setDomainAxis(xAxisCount, axis);
                plot.setDomainAxisLocation(xAxisCount, 
                        AxisLocation.TOP_OR_LEFT);
                axes.add(axis);
                xAxisCount++;
            }
            else if (c == ',') {
                // nothing to do
            }
            else {
                throw new RuntimeException("Bad character " + c);
            }
        }
        
    }

    /**
     * Processes a string that indicates the axes that should be visible on 
     * the plot.
     * 
     * @param plot  the plot.
     * @param axisStr  the axis specification.
     * @param axes  a list that will be populated with any axes added to the
     *              plot.
     */
    private void processAxisStrH(CategoryPlot plot, String axisStr, List axes) {
        int xAxisCount = 0;
        int yAxisCount = 0;
        for (int i = 0; i < axisStr.length(); i++) {
            char c = axisStr.charAt(i);
            if (c == 'y') {
                if (yAxisCount == 0) {
                    plot.getDomainAxis().setVisible(true);
                    axes.add(plot.getDomainAxis());
                    yAxisCount++;
                } 
                else {
                    GCategoryAxis axis = new GCategoryAxis();
                    axis.setAxisLineVisible(false);
                    plot.setDomainAxis(yAxisCount, axis);
                    plot.setDomainAxisLocation(xAxisCount, 
                            AxisLocation.BOTTOM_OR_LEFT);
                    axes.add(axis);
                    yAxisCount++;
                }
            }
            else if (c == 'x') {
                if (xAxisCount == 0) {
                    plot.getRangeAxis().setVisible(true);
                    axes.add(plot.getRangeAxis());
                    xAxisCount++;
                }
                else {
                    GValueAxis axis = new GValueAxis();
                    axis.setAxisLineVisible(false);
                    plot.setRangeAxis(xAxisCount, axis);
                    plot.setRangeAxisLocation(yAxisCount, 
                            AxisLocation.BOTTOM_OR_LEFT);
                    axes.add(axis);
                    xAxisCount++;
                }
            }
            else if (c == 't') {
                GValueAxis axis = new GValueAxis();
                plot.setRangeAxis(xAxisCount, axis);
                plot.setRangeAxisLocation(yAxisCount, 
                        AxisLocation.TOP_OR_LEFT);
                axes.add(axis);
                xAxisCount++;
            }
            else if (c == 'r') {
                GCategoryAxis axis = new GCategoryAxis();
                plot.setDomainAxis(yAxisCount, axis);
                plot.setDomainAxisLocation(xAxisCount, 
                        AxisLocation.BOTTOM_OR_RIGHT);
                axes.add(axis);
                yAxisCount++;
            }
            else if (c == ',') {
                // nothing to do
            }
            else {
                throw new RuntimeException("Bad character " + c);
            }
        }
        
    }

    /**
     * Parses a dataset from the incoming string.
     * 
     * @param dataStr  the data string.
     * 
     * @return A pie dataset.
     */
    private PieDataset parsePieDataset(String dataStr) {
        if (dataStr == null) {
            throw new IllegalArgumentException("Null 'dataStr' argument.");
        }
        DefaultPieDataset result = new DefaultPieDataset();
        List dataList = null;
        if (dataStr.startsWith("s:")) { // simple encoding
            dataList = parseSimpleData(dataStr);
        } 
        else if (dataStr.startsWith("t:")) {
            dataList = parseTextData(dataStr);
        }
        else if (dataStr.startsWith("e:")) {
            dataList = parseExtendedData(dataStr);
        }
        else {
            throw new RuntimeException("Unrecognised data encoding: " 
                    + dataStr.substring(0, 2));
        }
        
        if (dataList.size() > 0) {
            List series = (List) dataList.get(0);
            int s = 1;
            Iterator iterator = series.iterator();
            while (iterator.hasNext()) {
                Number i = (Number) iterator.next();
                result.setValue("S" + s, i);
                s++;
            }
        }
        else {
            throw new RuntimeException("Empty data list.");
        }
        return result;
    }
    
    /**
     * Parses a category dataset for use in bar charts.
     * 
     * @param dataStr  the data string.
     * @param yAxisRange  the y-axis range (used to scale the data values
     *         appropriately).
     *         
     * @return A category dataset.
     */
    private DefaultCategoryDataset parseCategoryDataset(String dataStr, 
            Range yAxisRange) {
        if (dataStr == null) {
            throw new IllegalArgumentException("Null 'dataStr' argument.");
        }
        List dataList = null;
        if (dataStr.startsWith("s:")) { // simple encoding
            dataList = parseSimpleData(dataStr);
        }
        else if (dataStr.startsWith("t:")) {
            dataList = parseTextData(dataStr);
        }
        else if (dataStr.startsWith("e:")) {
            dataList = parseExtendedData(dataStr);
        }
        else {
            throw new RuntimeException("Unrecognised data encoding: " 
                    + dataStr.substring(0, 2));
        }
        
        DefaultCategoryDataset result = new DefaultCategoryDataset();
        for (int i = 0; i < dataList.size(); i++) {
            List values = (List) dataList.get(i);
            for (int j = 0; j < values.size(); j++) {
                Number n = (Number) values.get(j);
                if (n != null) {
                    double y = yAxisRange.getLowerBound() + n.doubleValue() 
                            * yAxisRange.getLength();
                    result.setValue(y, String.valueOf(i), String.valueOf(j));
                }
                else {
                    result.setValue(n, String.valueOf(i), String.valueOf(j));                    
                }
            }
        }
        
        return result;
    }
    
    /**
     * Parses a dataset from the incoming string.  The data encoding only
     * specifies the y-values.
     * 
     * @param dataStr  the data string.
     * 
     * @return A pie dataset.
     */
    private XYDataset parseXYDataset(String dataStr, Range xAxisRange, 
            Range yAxisRange) {
        
        if (dataStr == null) {
            throw new IllegalArgumentException("Null 'dataStr' argument.");
        }
        List dataList;
        if (dataStr.startsWith("s:")) { // simple encoding
            dataList = parseSimpleData(dataStr);
        }
        else if (dataStr.startsWith("t:")) {
            dataList = parseTextData(dataStr);
        }
        else if (dataStr.startsWith("e:")) {
            dataList = parseExtendedData(dataStr);
        }
        else {
            throw new RuntimeException("Unrecognised data encoding: " 
                    + dataStr.substring(0, 2));
        }
        XYSeriesCollection result = new XYSeriesCollection();
        for (int s = 0; s < dataList.size(); s++) {
            List series = (List) dataList.get(s);
            XYSeries ss = new XYSeries("Series " + (s + 1));
            int itemCount = series.size();
            double increment = (xAxisRange.getLength()) / 
                    (Math.max(itemCount - 1, 1));
            for (int i = 0; i < itemCount; i++) {
                Number n = (Number) series.get(i);
                if (n != null) {
                    double v = n.doubleValue();
                    double y = yAxisRange.getLowerBound() 
                            + v * yAxisRange.getLength();
                    ss.add(xAxisRange.getLowerBound() + (i * increment), y);
                }
                else {
                    ss.add(xAxisRange.getLowerBound() + (i * increment), 
                            null);
                }
            }
            result.addSeries(ss);
        }
        return result;
    }
    
    /**
     * Parses a dataset from the incoming string.  The data encoding specifies 
     * both the x-values and the y-values, in consecutive series.
     * 
     * @param dataStr  the data string.
     * 
     * @return A pie dataset.
     */
    private XYDataset parseXYDataset2(String dataStr, Range xAxisRange, 
            Range yAxisRange) {
        
        if (dataStr == null) {
            throw new IllegalArgumentException("Null 'dataStr' argument.");
        }
        List dataList;
        if (dataStr.startsWith("s:")) { // simple encoding
            dataList = parseSimpleData(dataStr);
        }
        else if (dataStr.startsWith("t:")) {
            dataList = parseTextData(dataStr);
        }
        else if (dataStr.startsWith("e:")) {
            dataList = parseExtendedData(dataStr);
        }
        else {
            throw new RuntimeException("Unrecognised data encoding: " 
                    + dataStr.substring(0, 2));
        }
        XYSeriesCollection result = new XYSeriesCollection();
        for (int s = 0; s < dataList.size(); s = s + 2) {
            List xSeries = (List) dataList.get(s);
            List ySeries = (List) dataList.get(s + 1);
            XYSeries ss = new XYSeries("Series " + (s / 2));
            int itemCount = xSeries.size();
            // check for special case of x-values = -1
            boolean implicitXValues = false;
            if (itemCount == 1) {
                itemCount = ySeries.size();
                implicitXValues = true;
            }
            for (int i = 0; i < itemCount; i++) {
                double xx;
                if (implicitXValues) {
                    xx = xAxisRange.getLowerBound() + ((i + 0.0) / Math.max(
                            itemCount - 1, 1)) * (xAxisRange.getLength());
                }
                else {
                    Number x = (Number) xSeries.get(i);
                    xx = xAxisRange.getLowerBound() + x.doubleValue() 
                            * xAxisRange.getLength();
                }
                Number y = (Number) ySeries.get(i);
                if (y != null) {
                    double yy = yAxisRange.getLowerBound() 
                            + y.doubleValue() * yAxisRange.getLength();
                    ss.add(xx, yy);
                }
                else {
                    ss.add(xx, null);
                }
            }
            result.addSeries(ss);
        }
        return result;
    }
    
    /**
     * Parses a dataset from the incoming string.  The data encoding specifies 
     * both the x-values and the y-values, in consecutive series, with an 
     * optional third series supplying the relative marker sizes.
     * 
     * @param dataStr  the data string.
     * 
     * @return A pie dataset.
     */
    private XYSeriesCollection parseScatterDataset(String dataStr, 
            Range xAxisRange, Range yAxisRange) {
        
        if (dataStr == null) {
            throw new IllegalArgumentException("Null 'dataStr' argument.");
        }
        List dataList;
        if (dataStr.startsWith("s:")) { // simple encoding
            dataList = parseSimpleData(dataStr);
        }
        else if (dataStr.startsWith("t:")) {
            dataList = parseTextData(dataStr);
        }
        else if (dataStr.startsWith("e:")) {
            dataList = parseExtendedData(dataStr);
        }
        else {
            throw new RuntimeException("Unrecognised data encoding: " 
                    + dataStr.substring(0, 2));
        }
        XYSeriesCollection result = new XYSeriesCollection();
        List xSeries = (List) dataList.get(0);
        List ySeries = (List) dataList.get(1);
        XYSeries ss = new XYSeries("Series");
        int itemCount = xSeries.size();
       
        for (int i = 0; i < itemCount; i++) {
            Number x = (Number) xSeries.get(i);
            double xx = xAxisRange.getLowerBound() + x.doubleValue() 
                    * xAxisRange.getLength();
            Number y = (Number) ySeries.get(i);
            if (y != null) {
                double yy = yAxisRange.getLowerBound() 
                        + y.doubleValue() * yAxisRange.getLength();
                ss.add(xx, yy);
            }
            else {
                ss.add(xx, null);
            }
            result.addSeries(ss);
        }
        
        // is there a third series?
        List zSeries = null;
        if (dataList.size() > 2) {
            zSeries = (List) dataList.get(2);
            XYSeries sz = new XYSeries("Z");
            for (int i = 0; i < zSeries.size(); i++) {
                sz.add(i, (Number) zSeries.get(i));
            }
        }

        return result;
    }
        
    /**
     * Parses a string with simple data encoding.  The method returns a list
     * of lists, with the sublists containing one Double object per data
     * value (in the range 0.0 to 1.0).
     * 
     * @param data  the data string.
     * 
     * @return A list containing a sublist of Integer objects for each series.
     */
    private List parseSimpleData(String data) {
        List dataset = new java.util.ArrayList(); 
        StringCharacterIterator iterator = new StringCharacterIterator(data, 2);
        List series = new java.util.ArrayList();
        while (iterator.current() != CharacterIterator.DONE) {
            char c = iterator.current();
            if (c == ',') {
                dataset.add(series);
                series = new java.util.ArrayList();
            }
            else {
                int v = googleValue(c);
                if (v < Integer.MAX_VALUE) {
                    series.add(new Double(v / 61.0));
                }
                else {
                    series.add(null);
                }
            }
            iterator.next();
        }
        dataset.add(series);
        return dataset;
    }
    
    /**
     * Parses a string with text data encoding.
     * 
     * @param dataStr  the data string.
     * 
     * @return A list containing a sublist of Float objects for each series.
     */
    private List parseTextData(String dataStr) {
        List dataset = new java.util.ArrayList(); 
        StringCharacterIterator iterator = new StringCharacterIterator(dataStr,
                2);
        List series = new java.util.ArrayList();
        int baseIndex = 2;
        while (iterator.current() != CharacterIterator.DONE) {
            char c = iterator.current();
            if (c >= '0' && c <= '9' || c == '-' || c == '.') {
                iterator.next();
            }
            else if (c == ',') {
                series.add(new Float(Float.parseFloat(dataStr.substring(
                        baseIndex, iterator.getIndex())) / 100.0f));
                baseIndex = iterator.getIndex() + 1;
                iterator.next();
            }
            else if (c == '|') {
                series.add(new Float(Float.parseFloat(dataStr.substring(
                        baseIndex, iterator.getIndex())) / 100.0f));
                baseIndex = iterator.getIndex() + 1;
                dataset.add(series);
                series = new java.util.ArrayList();
                iterator.next();
            }
            else {
                throw new IllegalArgumentException(
                        "'data' contains bad characters.");
            }
        }
        series.add(new Float(Float.parseFloat(dataStr.substring(
                baseIndex, iterator.getIndex())) / 100.0f));
        dataset.add(series);
        return dataset;
    }
    
    /**
     * Parses a string with simple data encoding.  The method returns a list
     * of lists, with the sublists containing one Double object per data
     * value (in the range 0.0 to 1.0).
     * 
     * @param data  the data string.
     * 
     * @return A list containing a sublist of Number objects for each series.
     */
    private List parseExtendedData(String data) {
        List dataset = new java.util.ArrayList(); 
        StringCharacterIterator iterator = new StringCharacterIterator(data, 2);
        List series = new java.util.ArrayList();
        while (iterator.current() != CharacterIterator.DONE) {
            char c1 = iterator.current();
            if (c1 == ',') {
                dataset.add(series);
                series = new java.util.ArrayList();
            }
            else {
                char c2 = iterator.next();
                int c1v = googleValueExt(c1);
                int c2v = googleValueExt(c2);
                if (c1v < Integer.MAX_VALUE && c2v < Integer.MAX_VALUE) {
                    series.add(new Double((c1v * 64.0 + c2v) / 4095.0));
                }
                else {
                    series.add(null);
                }
            }
            iterator.next();                               
        }
        dataset.add(series);
        return dataset;
    }
    
    /**
     * Returns the value of the specified character in the Google Charts API 
     * simple encoding.
     * 
     * @param c  the character.
     * 
     * @return The value.
     */
    private int googleValue(char c) {
        if (c == '_') {
            return Integer.MAX_VALUE;
        }
        else if (c >= 'a' && c <= 'z') {
            return c - 97 + 26;    
        }
        else if (c >= 'A' && c <= 'Z') {
            return c - 65;                        
        }
        else if (c >= '0' && c <= '9') {
            return c - 48 + 52;                        
        }
        else {
            throw new IllegalArgumentException("Invalid character '" 
                    + c + "'.");
        }
    }
    
    /**
     * Returns the value of the specified character in the Google Charts API 
     * extended encoding.
     * 
     * @param c  the character.
     * 
     * @return The value.
     */
    private int googleValueExt(char c) {
        if (c == '_') {
            return Integer.MAX_VALUE;
        }
        else if (c >= 'a' && c <= 'z') {
            return c - 97 + 26;    
        }
        else if (c >= 'A' && c <= 'Z') {
            return c - 65;                        
        }
        else if (c >= '0' && c <= '9') {
            return c - 48 + 52;                        
        }
        else if (c == '-') {
            return 62;
        }
        else if (c == '.') {
            return 63;
        }
        else {
            throw new IllegalArgumentException("Invalid character '" 
                    + c + "'.");
        }
    }

    /**
     * Parses a string containing the chart dimensions.
     * 
     * @param text  the text (<code>null</code> not permitted).
     * 
     * @return The chart dimensions.
     * 
     * @throws ServletException
     */
    private int[] parseChartDimensions(String text) throws ServletException {
        if (text == null) {
            throw new IllegalArgumentException(
                    "Null 'text' argument (in parseChartDimensions(String)).");
        }
        int[] result = new int[2];
        int splitIndex = text.indexOf('x');
        String xStr = text.substring(0, splitIndex);
        String yStr = text.substring(splitIndex + 1);
        int x = Integer.parseInt(xStr);
        int y = Integer.parseInt(yStr);
        if (x <= 1000 && y <= 1000 && x * y < 360000) {
            result[0] = x; result[1] = y;
        }
        else {
            throw new ServletException("Invalid chart dimensions: " + xStr 
                    + ", " + yStr);
        }
        return result;
    }

    /**
     * Creates a pie chart.
     *
     * @return A pie chart.
     */
    private JFreeChart createPieChart() {
        JFreeChart chart = ChartFactory.createPieChart(null, null, 
                false, false, false);
        chart.setBackgroundPaint(Color.white);
        PiePlot plot = (PiePlot) chart.getPlot();
        plot.setInsets(RectangleInsets.ZERO_INSETS);
        plot.setInteriorGap(0.06);
        plot.setStartAngle(0.0);
        plot.setLabelGenerator(null);
        plot.setBaseSectionOutlinePaint(Color.white);
        plot.setBaseSectionOutlineStroke(new BasicStroke(1.2f, 
                BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
        plot.setOutlineVisible(false);
        plot.setLabelBackgroundPaint(Color.white);
        plot.setLabelOutlinePaint(null);
        plot.setLabelShadowPaint(null);
        plot.setLabelPadding(RectangleInsets.ZERO_INSETS);
        plot.setLabelFont(new Font("Dialog", Font.PLAIN, 12));
        plot.setLabelPaint(Color.gray);
        return chart;
    }

    /**
     * Creates a pie chart with 3D effect.
     *
     * @return A pie chart.
     */
    private JFreeChart createPieChart3D() {
        JFreeChart chart = ChartFactory.createPieChart3D(null, null, 
                false, false, false);
        chart.setBackgroundPaint(Color.white);
        chart.setBorderPaint(Color.white);
        PiePlot3D plot = (PiePlot3D) chart.getPlot();
        plot.setInsets(RectangleInsets.ZERO_INSETS);
        plot.setDarkerSides(true);
        plot.setBaseSectionOutlinePaint(new Color(0,0,0,0));
        plot.setStartAngle(0.0);
        plot.setInteriorGap(0.10);
        plot.setLabelGenerator(null);
        plot.setOutlineVisible(false);
        plot.setLabelBackgroundPaint(Color.white);
        plot.setLabelOutlinePaint(null);
        plot.setLabelShadowPaint(null);
        plot.setLabelPadding(RectangleInsets.ZERO_INSETS);
        plot.setLabelFont(new Font("Dialog", Font.PLAIN, 12));
        plot.setLabelPaint(Color.gray);
        return chart;
    }

    /**
     * Creates a line chart.
     * 
     * @return A line chart.
     */
    private JFreeChart createLineChart() {
        GXYPlot plot = new GXYPlot();
        XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer();
        renderer.setBaseShapesVisible(false);
        plot.setRenderer(renderer);
        JFreeChart chart = new JFreeChart(plot);
        chart.removeLegend();
        chart.setBackgroundPaint(Color.white);
        renderer.setBasePaint(new Color(0xFF9900));
        renderer.setBaseStroke(new BasicStroke(1.5f, BasicStroke.CAP_ROUND,
                BasicStroke.JOIN_ROUND));
        renderer.setAutoPopulateSeriesPaint(false);
        
        GValueAxis xAxis = new GValueAxis();
        xAxis.setVisible(false);
        plot.setDomainAxis(xAxis);
        GValueAxis yAxis = new GValueAxis();
        yAxis.setVisible(false);
        plot.setRangeAxis(yAxis);
        plot.setDomainGridlinesVisible(false);
        plot.setRangeGridlinesVisible(false);
        return chart;
    }
    
    /**
     * Creates a bar chart with the specified orientation.
     * 
     * @param orientation  the plot orientation.
     * 
     * @return A bar chart.
     */
    private JFreeChart createBarChart(PlotOrientation orientation) {
        JFreeChart chart = ChartFactory.createBarChart(null, null, null, null, 
                orientation, false, false, false);
        chart.setBackgroundPaint(Color.white);
        CategoryPlot plot = (CategoryPlot) chart.getPlot();
        plot.setBackgroundPaint(null);
        BarRenderer renderer = new BarRenderer();
        plot.setRenderer(renderer);
        renderer.setBasePaint(new Color(0xFFCC33));
        renderer.setAutoPopulateSeriesPaint(false);
        GCategoryAxis xAxis = new GCategoryAxis();
        xAxis.setVisible(false);
        xAxis.setMaximumCategoryLabelLines(5);
        plot.setDomainAxis(xAxis);
        GValueAxis yAxis = new GValueAxis();
        yAxis.setVisible(false);
        plot.setRangeAxis(yAxis);
        plot.setRangeAxisLocation(AxisLocation.BOTTOM_OR_LEFT);
        plot.setDomainGridlinesVisible(false);
        plot.setRangeGridlinesVisible(false);
        return chart;
    }

    /**
     * Creates a stacked bar chart with the specified orientation.
     * 
     * @param orientation  the orientation.
     * 
     * @return A stacked bar chart.
     */
    private JFreeChart createStackedBarChart(PlotOrientation orientation) {
        JFreeChart chart = ChartFactory.createStackedBarChart(null, null, 
                null, null, orientation, false, false, false);
        chart.setBackgroundPaint(Color.white);
        CategoryPlot plot = (CategoryPlot) chart.getPlot();
        plot.setBackgroundPaint(null);
        StackedBarRenderer renderer = new StackedBarRenderer();
        renderer.setBasePaint(new Color(0xFFCC33));
        renderer.setAutoPopulateSeriesPaint(false);
        plot.setRenderer(renderer);
        GCategoryAxis xAxis = new GCategoryAxis();
        xAxis.setVisible(false);
        xAxis.setMaximumCategoryLabelLines(5);
        plot.setDomainAxis(xAxis);
        GValueAxis yAxis = new GValueAxis();
        yAxis.setVisible(false);
        plot.setRangeAxis(yAxis);
        plot.setRangeAxisLocation(AxisLocation.BOTTOM_OR_LEFT);
        plot.setDomainGridlinesVisible(false);
        plot.setRangeGridlinesVisible(false);
        return chart;
    }

    /**
     * Creates a scatter chart.
     * 
     * @return A scatter chart.
     */
    private JFreeChart createScatterChart() {
        JFreeChart chart = ChartFactory.createScatterPlot(null, null, null, 
                null, PlotOrientation.VERTICAL, false, false, false);
        chart.setBackgroundPaint(Color.white);
        XYPlot plot = (XYPlot) chart.getPlot();
        plot.setBackgroundPaint(null);
        plot.setOutlinePaint(null);
        XYLineAndShapeRenderer renderer 
                = (XYLineAndShapeRenderer) plot.getRenderer();
        renderer.setBasePaint(new Color(0x76A4FB));
        renderer.setAutoPopulateSeriesPaint(false);
        
        GValueAxis xAxis = new GValueAxis();
        xAxis.setVisible(false);
        plot.setDomainAxis(xAxis);
        GValueAxis yAxis = new GValueAxis();
        yAxis.setVisible(false);
        plot.setRangeAxis(yAxis);
        plot.setDomainGridlinesVisible(false);
        plot.setRangeGridlinesVisible(false);
        return chart;
    }
    
}

