/* ======================
 * Eastwood Chart Servlet
 * ======================
 *
 * (C) Copyright 2007, by Object Refinery Limited.
 *
 * Project Info:  http://www.jfree.org/eastwood/index.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc. 
 * in the United States and other countries.]
 *
 * ---------------
 * GValueAxis.java
 * ---------------
 * (C) Copyright 2007, by Object Refinery Limited.
 *
 * Original Author:  David Gilbert (for Object Refinery Limited);
 *
 * Changes
 * -------
 * 13-Dec-2007 : Version 1 (DG);
 *
 */

package org.jfree.eastwood;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.List;

import org.jfree.chart.axis.AxisState;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTick;
import org.jfree.chart.axis.Tick;
import org.jfree.data.Range;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.TextAnchor;

/**
 * An axis implementation designed exclusively for use with the 
 * {@link ChartServlet} servlet.
 */
class GValueAxis extends NumberAxis implements GLabelledAxis {

    /** The tick labels. */
    private List tickLabels;
    
    /** The tick values (as Number objects). */
    private List tickValues;
    
    /**
     * Creates a new axis with the default range 0 to 100.  
     */
    public GValueAxis() {
        super();
        this.tickLabels = new java.util.ArrayList();
        this.tickValues = new java.util.ArrayList();
        setLowerMargin(0.0);
        setUpperMargin(0.0);
        setRange(0, 100);
        setStandardTickUnits(NumberAxis.createIntegerTickUnits());
        setTickLabelPaint(Color.gray);
        setTickLabelFont(new Font("Dialog", Font.PLAIN, 11));
        setTickMarksVisible(false);
    }

    /**
     * Sets the labels to be displayed along the axis.
     * 
     * @param labels  the labels (<code>null</code> not permitted).
     */
    public void setTickLabels(List labels) {
        if (labels == null) {
            throw new IllegalArgumentException("Null 'labels' argument.");
        }
        this.tickLabels = labels;
        setAutoTickUnitSelection(false);
    }
    
    /**
     * Sets the tick values.
     * 
     * @param values  the tick values (<code>null</code> not permitted.
     */
    public void setTickValues(List values) {
        if (values == null) {
            throw new IllegalArgumentException("Null 'values' argument.");
        }
        this.tickValues = values;
        setAutoTickUnitSelection(false);
    }

    /**
     * Creates the list of ticks to display.
     * 
     * @param g2  the graphics target.
     * @param state  
     * @param dataArea
     * @param edge
     * 
     * @return The list.
     */
    public List refreshTicks(Graphics2D g2, AxisState state, 
            Rectangle2D dataArea, RectangleEdge edge) {
        
        if (isAutoTickUnitSelection()) {
            return super.refreshTicks(g2, state, dataArea, edge);
        }
        
        List result = new java.util.ArrayList();
        
        Range range = getRange();
        double axisMin = range.getLowerBound();
        double axisMax = range.getUpperBound();
        double length = axisMax - axisMin;
        int labelCount = this.tickLabels.size();
        int valueCount = this.tickValues.size();
        int tickCount = Math.max(labelCount, valueCount);
        
        // work out the label anchor points according to which side of the
        // chart the axis lies on...
        TextAnchor anchor = null;
        TextAnchor rotationAnchor = TextAnchor.CENTER;
        if (edge == RectangleEdge.LEFT) {
            anchor = TextAnchor.CENTER_RIGHT;
        }
        else if (edge == RectangleEdge.BOTTOM) {
            anchor = TextAnchor.TOP_CENTER;         
        }
        else if (edge == RectangleEdge.TOP) {
            anchor = TextAnchor.BOTTOM_CENTER;                     
        }
        else if (edge == RectangleEdge.RIGHT) {
            anchor = TextAnchor.CENTER_LEFT;
        }
        
        for (int i = 0; i < tickCount; i++) {
            String tickLabel = null;
            if (i < labelCount) {
                tickLabel = this.tickLabels.get(i).toString();
            }
            else {
                tickLabel = String.valueOf(this.tickValues.get(i));
            }
            
            // is there a value specified
            double tickValue = Double.NaN;
            if (i < valueCount) {
                Number tv = (Number) this.tickValues.get(i);
                if (tv != null) {
                    tickValue = tv.doubleValue();
                }
            }
            if (Double.isNaN(tickValue)) {
                tickValue = axisMin + (i * (length / (Math.max(labelCount - 1, 
                        1))));
            }

            Tick tick = new NumberTick(new Double(tickValue), tickLabel, 
                    anchor, rotationAnchor, 0.0);
            
            result.add(tick);
        }
        return result;
    }

}
